

{-----------------------------------------------------------------------------
 Unit Name: FImplicitCurve
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/01/02 
 Purpose: class TImplicitCurve, that implements an algorithm to solve an
 equation in two variables, that is in the form f(x,y)=0. 
-----------------------------------------------------------------------------}
{*-
 * Copyright (c) 2002 Angel Montesinos Amilibia
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *        Version 4.0.0.0
 *}



unit FImplicitCurve;

interface

uses
  SysUtils, Classes
{$IFDEF MSWINDOWS}
  , Windows, Forms, Graphics,
{$ENDIF}
{$IFDEF LINUX}
  , QGraphics, QForms, QTypes,
{$ENDIF}
  Globals;

type
  RecordAndDrawProcedure2D = procedure(previousX, currentX: Vector2; key:
    Integer);
  CheckRange2DFunction = function(var depth: Double; p: Vector2): Boolean;
  R2ToR1Procedure = procedure(var f: double; p: Vector2);

  TImplicitCurve = class(TObject)
  private
    theF: R2ToR1Procedure;
    InRange: CheckRange2DFunction;
    RecordAndDraw: RecordAndDrawProcedure2D;
    minCorner, maxCorner, step, flagStep, halfEdges, center, incr: Vector2;
    coordVert: array[1..2, 0..kMaxPPLine2D] of double;
                         {Coordinates of the  bottom-left}
                         {corners of the rectangles.     }
    ppLine, flagsPLine: Integer;
    verySmallF, maxLength, minLength, minShift, radius: Double;

    alreadyUsed: packed array[0..kMaxNumFlags, 0..kMaxNumFlags] of boolean;

    procedure SetEnvironment(theF0: R2ToR1Procedure; InRange0:
      CheckRange2DFunction;
      RecordAndDraw0: RecordAndDrawProcedure2D; verySmallF0: Double;
      minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
      limitSegments0: Boolean);
    function GetNearestZero(thePoint, dirSignaller: Vector2; natural:
      Boolean; var
      advance: double; var theZero: Vector2): Boolean;
    function GetFirstPair(starter: Vector2; var leaders: Segment2): Integer;
    function MarkAsUsed(iIni, jIni, iFini, jFini: Integer): Boolean;
    procedure FollowLine(ini: Vector2; u1, u2: Integer);
    function InitialPointInRectangle(ci, cj: Integer;
      var startingPoint: Vector2): Boolean;
  public
    procedure ComputeImplicitCurve(theF0: R2ToR1Procedure; InRange0:
      CheckRange2DFunction;
      RecordAndDraw0: RecordAndDrawProcedure2D; verySmallF0: Double;
      minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
      limitSegments0: Boolean);
  end;

var
  theImplicitCurve: TImplicitCurve;

implementation

uses
  Math, Messages,
  FSurface;
var
  coordRatios: array[1..3] of double;
  cosAndSin: array[1..2, 0..15] of double;

procedure TImplicitCurve.SetEnvironment(theF0: R2ToR1Procedure;
  InRange0: CheckRange2DFunction;
  RecordAndDraw0: RecordAndDrawProcedure2D; verySmallF0: Double;
  minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
  limitSegments0: Boolean);
var
  i, j: Integer;
begin
  theF := theF0;
  InRange := InRange0;
  RecordAndDraw := RecordAndDraw0;
  verySmallF := verySmallF0;
  minCorner := minCorner0;
  maxCorner := maxCorner0;
  ppLine := ppLine0;
  flagsPLine := flagsPLine0;
  for i := 1 to 2 do begin
    center[i] := (maxCorner[i] + minCorner[i]) / 2;
    halfEdges[i] := (maxCorner[i] - minCorner[i]) / 2;
    step[i] := (maxCorner[i] - minCorner[i]) / ppLine;
    flagStep[i] := (maxCorner[i] - minCorner[i]) / flagsPLine;
    incr[i] := step[i] / 128;
    coordVert[i, 0] := minCorner[i];
    for j := 1 to ppLine do
      coordVert[i, j] := coordVert[i, j - 1] + step[i];
  end;
  maxLength := 0.7 * hypot(step[1], step[2]);
  if limitSegments0 then
    minLength := maxLength
  else
    minLength := 0;
  radius := hypot(halfEdges[1], halfEdges[2]) / 200;
  minShift := radius / 100;
end;

function TImplicitCurve.GetNearestZero(thePoint, dirSignaller: Vector2;
  natural: Boolean; var
  advance: double; var theZero: Vector2): Boolean;
{A Newton process is applied recursively starting from 'thePoint' to get          }
{a zero of 'theF'. If 'natural' is TRUE, the direction of search is that given by }
{grad(theF). If not, the direction is made orthogonal to the vector that connects }
{'thePoint' with 'dirSignaller'. The process is assumed to succeed if the value of}
{'theF' at the last point is less that 'verySmallF; then 'GetNearestZero'     }
{results in TRUE. Otherwise, in FALSE. In 'advance' it is given approximately the }
{real number such that theZero:= thePoint+ advance*grad(theF) in the natural case,}
{or that theZero:=thePoint+advance*J(dirSignaller-thePoint), where J gives the    }
{pi/2 positive rotation of a vector.                                              }

const
  cMinDenom: double = 1.0E-10;
var
  moving, moved, gradient, director, shift: Vector2;
  f0, f2, f1, f3, f4, aux, theStep, previousStep: double;
  badPoint, veryNear, found: Boolean;
  i: Integer;
begin
  Result := FALSE;
  if not natural then begin
    director[1] := -dirSignaller[2] + thePoint[2];
    director[2] := dirSignaller[1] - thePoint[1];
    if hypot(director[1], director[2]) < cMinDenom then
      exit;
  end;
  moving := thePoint;
  moved := thePoint;
  advance := 0.0;
  theF(f0, moving);
  if abs(f0) < verySmallF then begin
    GetNearestZero := TRUE;
    theZero := moving;
    exit;
  end;
  theStep := 0.2 * (step[1] + step[2]);

  repeat
    previousStep := theStep;
    moved[1] := moving[1] - 1.5*incr[1];
    moved[2] := moving[2];
    theF(f1, moved);
    moved[1] := moving[1] - incr[1]/2;
    theF(f2, moved);
    moved[1] := moving[1] + incr[1]/2;
    theF(f3, moved);
    moved[1] := moving[1] + 1.5*incr[1];
    theF(f4, moved);
    gradient[1] := (f1 + 27 * (f3 - f2) - f4) / (24*incr[1]);
    moved[1] := moving[1] ;
    moved[2] := moving[2]- 1.5*incr[2];
    theF(f1, moved);
    moved[2] := moving[2] - incr[2]/2;
    theF(f2, moved);
    moved[2] := moving[2] + incr[2]/2;
    theF(f3, moved);
    moved[2] := moving[2] + 1.5*incr[2];
    theF(f4, moved);
    gradient[2] := (f1 + 27 * (f3 - f2) - f4) / (24*incr[2]);
    if natural then
      aux := sqr(gradient[1]) + sqr(gradient[2])
    else
      aux := gradient[1] * director[1] + gradient[2] * director[2];
    if abs(aux) < cMinDenom then
      exit;
    aux := -f0 / aux;
    if natural then
      for i := 1 to 2 do begin
        shift[i] := aux * gradient[i];
        moving[i] := moving[i] + shift[i];
      end
    else
      for i := 1 to 2 do begin
        shift[i] := aux * director[i];
        moving[i] := moving[i] + shift[i];
      end;
    theStep := abs(shift[1]) + abs(shift[2]);
    theF(f0, moving);
    badPoint := (theStep > 0.9 * previousStep);
    advance := advance + aux;
    found := (abs(f0) < verySmallF);
    veryNear := {(theStep < minShift) or} found;
  until veryNear or badPoint;
  if found then begin
    theZero := moving;
    Result := True;
  end;
end; { of GetNearestZero}

function TImplicitCurve.GetFirstPair(starter: Vector2; var leaders:
  Segment2): Integer;
{It is assumed that the absolute value of f(starter) is less than              }
{'verySmallF'. The function tries to determine two points placed symmetrically  }
{around 'starter',that will be given in 'leaders', on which the absolute value    }
{of 'f' is also less than 'verySmallF'. The distance from 'starter' to each  }
{point of 'leaders' is approximately equal to 'radius'. The function  results  }
{in the number of points so obtained.           }

var
  i, j, theBest: Integer;
  positions: array[0..15] of Vector2;
  badness, value, distance, advance: double;
  success: Boolean;
begin
  result := 0;
  badness := 1.0E100;
  theBest := 0;

  {First we look for the best point in the circle of radius 'radius' around 'starter'}
  for i := 0 to 15 do begin
    for j := 1 to 2 do
      positions[i][j] := starter[j] + radius * cosAndSin[j, i];
    theF(value, positions[i]);
    value := abs(value);
    if value < badness then begin
      badness := value;
      theBest := i;
    end;
  end; 
  success := GetNearestZero(positions[theBest], starter, FALSE, advance,
    leaders[1]);
  if success then begin
    distance := hypot(starter[1] - leaders[1][1], starter[2] - leaders[1][2]);
    success := (distance < 2.0 * radius);
  end;
  if success then
    result := 1;

 {Now, we proceed to the second point in a direction opposite to the first one.   }
  success := GetNearestZero(positions[(theBest + 8) mod 16], starter, FALSE,
    advance, leaders[2]);
  if success then begin
    distance := hypot(starter[1] - leaders[2][1], starter[2] - leaders[2][2]);
    success := (distance < 3.0 * radius)
  end;
  if not success then
    exit
  else begin
    if result = 1 then
      inc(result)
    else begin
      result := 1;
      leaders[1] := leaders[2];
    end;
  end;
end; { of GetFirstPair}



function TImplicitCurve.MarkAsUsed(iIni, jIni, iFini, jFini: Integer):
  Boolean;
{If a segment in the domain goes from the small flag rectangle identified by      }
{'iIni, jIni', to the one identified by 'iFini,jFini', both rectangles and        }
{the intermediate ones are marked as used. If one of them (not the first)         }
{was already used, the function returns TRUE; else, it returns FALSE.             }
var
  pitch: double;
  height, k, l, increment: Integer;
begin
  Result := False;
  if (iIni < 0) or (jIni < 0) or (iFini < 0) or (jFini < 0) then
    Exit;
  if (iIni > kMaxNumFlags) or (jIni > kMaxNumFlags) or (iFini >
    kMaxNumFlags) or (jFini > kMaxNumFlags) then
    Exit;
  markAsUsed := FALSE;
  alreadyUsed[iIni, jIni] := TRUE;
  if (iFini <> iIni) or (jFini <> jIni) then begin
    if abs(iFini - iIni) >= abs(jFini - jIni) then begin
      if iFini > iIni then
        increment := 1
      else
        increment := -1;
      pitch := (jFini - jIni) / (iFini - iIni);
      k := 0;
      repeat
        k := k + increment;
        l := iIni + k;
        height := round(jIni + k * pitch);
        if alreadyUsed[l, height] then
          markAsUsed := TRUE
        else
          alreadyUsed[l, height] := TRUE;
      until (l = iFini);
    end
    else begin
      if jFini > jIni then
        increment := 1
      else
        increment := -1;
      pitch := (iFini - iIni) / (jFini - jIni);
      k := 0;
      repeat
        k := k + increment;
        l := jIni + k;
        height := round(iIni + k * pitch);
        if alreadyUsed[height, l] then
          markAsUsed := TRUE
        else
          alreadyUsed[height, l] := TRUE;
      until (l = jFini);
    end
  end;
end;




procedure TImplicitCurve.FollowLine(ini: Vector2; u1, u2:
  Integer);
{It is supposed that 'ini' is a point in the subset where 'f' vanishes. First, }
{we apply 'GetFirstPair' to obtain two points in that subset that signal two    }
{marching directions. We apply a marching algorithm, whose essence is             }
{'GetNextPoint', in both directions.                                            }
const
  allowForRoundingErrors: double = 1.0E-8;
var
  leaders, elSegmento: Segment2;
  i, j, iCurrent, jCurrent, iFormer, jFormer, numStarters: Integer;
  alpha0, alpha1, alpha2, alpha3: Vector2;
  value, depth, formerDepth, initialDepth, advance: double;
  rectChange, goOn, goesOut, repeated: Boolean;


  function GetNextPoint: Boolean;
 {Assume that the points 'alpha0, alpha1, alpha2' are in the curve where 'f'    }
 {vanishes. We find, in a parabola passing by those three points, a new point 'z', }
 {such that the tangent to the parabola at 'z' and that tangent at 'alpha2' form an}
 {angle whose tan is equal to 'tanA'. Corrections are applied to 'z' if it happens }
 {to be placed too far, or too close with respect to 'alpha2'. Then, we apply      }
 {'GetNearestZero' to 'z' to obtain the new point 'alpha3'.                      }
  const
    tanA: double = 0.08;
  var
    count, i: Integer;
    y, Jy, z: Vector2;
    goodResult: Boolean;
    a, b, c, r, rPlusB, aTanA, t, denom, coef: double;
  begin
    GetNextPoint := FALSE;
    b := hypot(alpha2[1] - alpha1[1], alpha2[2] - alpha1[2]);
    if b < 1.0E-8 then
      exit;
    for i := 1 to 2 do
      y[i] := (alpha2[i] - alpha1[i]) / b;
    r := (alpha0[1] - alpha2[1]) * y[1] + (alpha0[2] - alpha2[2]) * y[2];
    c := sqr(alpha2[1] - alpha0[1]) + sqr(alpha2[2] - alpha0[2]) - sqr(r);
    if c > 1.0E-7 then begin
      rPlusB := r + b;
      if (rPlusB > -1.0E-6) then
        exit;
      c := sqrt(c);
      for i := 1 to 2 do
        Jy[i] := (alpha0[i] - alpha2[i] - r * y[i]) / c;
      a := ((alpha0[1] - alpha2[1]) * Jy[1] + (alpha0[2] -
        alpha2[2]) * Jy[2]) / (r * rPlusB);
      aTanA := abs(a * tanA);
      denom := 2.0 * (1 - b * aTanA);
      if (aTanA < 1.0E-8) or (denom < 1.0E-8) then
        t := 1.5 * maxLength
      else begin
        t := (sqr(b) + sqr(1 / a)) * aTanA / denom;
        t := min(t, maxLength);
      end;
    end
    else {that is, 0 <=  c <= 1.0e-7} begin
      a := 0;
      t := 1.5 * maxLength;
      Jy[1] := 0;
      Jy[2] := 0;
    end;
    if t < 0.1 * maxLength then
      t := maxLength / 8;
    coef := a * t * (t + b);
    for i := 1 to 2 do
      z[i] := alpha2[i] + t * y[i] + coef * Jy[i];
    count := 0;
    repeat
      count := count + 1;
      goodResult := GetNearestZero(z, alpha2, FALSE, c, alpha3);
      goodResult := goodResult and (c < tanA);
      if not goodResult then begin
        t := 0.5 * t;
        coef := a * t * (t + b);
        for i := 1 to 2 do
          z[i] := alpha2[i] + t * y[i] + coef * Jy[i];
      end;
    until goodResult or (count = 5);
    if goodResult then
      GetNextPoint := TRUE;
  end;

begin {of  FollowLine}
  value := 0;
  advance := 0;
  numStarters := GetFirstPair(ini, leaders);

  if numStarters = 0 then
    exit;
  InRange(initialDepth, ini);
  for i := 1 to numStarters do begin
    RecordAndDraw(ini, ini, 1);
    repeated := FALSE;
    goOn := TRUE;
    if not InRange(depth, leaders[i]) or (depth < 0) then begin
      value := 0;
      if depth < 0 then
        value := (initialDepth - allowForRoundingErrors) / (initialDepth +
          abs(depth));
      for j := 1 to 2 do
        leaders[i][j] := ini[j] + value * (leaders[i][j] - ini[j]);
      RecordAndDraw(ini, leaders[i], 0);
      break;
    end;

    iCurrent := Floor((leaders[i][1] - minCorner[1]) / flagStep[1]);
    jCurrent := Floor((leaders[i][2] - minCorner[2]) / flagStep[2]);
    elSegmento[2] := leaders[i];

    RecordAndDraw(ini, elSegmento[2], 0);
    rectChange := (iCurrent <> u1) or (jCurrent <> u2);
    if rectChange then begin
      if markAsUsed(u1, u2, iCurrent, jCurrent) then
        repeated := TRUE
      else
        repeated := FALSE;
      goOn := (not repeated);
    end;
    if not goOn then begin
      break;
    end;
    if numStarters = 2 then begin
      alpha0 := leaders[3 - i];
      alpha1 := ini;
      alpha2 := leaders[i];
      goOn := GetNextPoint;
    end
    else begin
      alpha1 := ini;
      alpha2 := leaders[1];
      for j := 1 to 2 do
        alpha0[j] := 2.0 * alpha2[j] - alpha1[j];
      goOn := GetNearestZero(alpha0, alpha2, FALSE, value, alpha3);
      goOn := goOn and (abs(value) < 0.1);
    end; {of  IF numStarters=2... ELSE}
    if not goOn then
      break;

    repeat
      elSegmento[1] := elSegmento[2];
      goesOut := FALSE;
      formerDepth := depth;
      if not InRange(depth, alpha3) or (depth < 0) then begin
        value := 0;
        if depth < 0 then
          value := (formerDepth - allowForRoundingErrors) / (formerDepth +
            abs(depth));
        for j := 1 to 2 do
          alpha3[j] := alpha2[j] + value * (alpha3[j] - alpha2[j]);
        goesOut := TRUE;
        goOn := False;
      end;
      iformer := iCurrent;
      jFormer := jCurrent;
      iCurrent := Floor((alpha3[1] - minCorner[1]) / flagStep[1]);
      jCurrent := Floor((alpha3[2] - minCorner[2]) / flagStep[2]);

      rectChange := (iCurrent <> iFormer) or (jCurrent <> jFormer);
      if rectChange and (not goesOut) then begin
        if markAsUsed(iFormer, jFormer, iCurrent, jCurrent) then
          repeated := TRUE
        else
          repeated := FALSE;
        goOn := (not repeated) and (not goesOut);
      end;
      elSegmento[2] := alpha3;
      advance := advance + hypot(elSegmento[2][1] - elSegmento[1][1],
        elSegmento[2][2] - elSegmento[1][2]);
      if (advance > minLength) or (not goOn) then begin
        RecordAndDraw(elSegmento[1], elSegmento[2], 0);
        advance := 0;
      end;

      if not goOn then
        break;
      alpha0 := alpha1;
      alpha1 := alpha2;
      alpha2 := alpha3;
      goOn := GetNextPoint;
    until (not goOn) or repeated or goesOut;
    if advance > minLength then
      RecordAndDraw(elSegmento[1], elSegmento[2], 0);
    if (i = 1) and (numStarters = 2) then
      alreadyUsed[u1, u2] := FALSE;
  end; {of FOR i:=1 TO numStarters DO}
end; {of  FollowLine}



procedure InitializeTImplicitCurve;
var
  i: Integer;
  z: double;
begin
  theImplicitCurve := TImplicitCurve.Create;
  coordRatios[1] := 1 / 6;
  coordRatios[2] := 3 / 6;
  coordRatios[3] := 5 / 6;
  for i := 0 to 15 do begin
    z := i * pi / 8;
    cosAndSin[1, i] := cos(z);
    cosAndSin[2, i] := sin(z);
  end;
end;

function TImplicitCurve.InitialPointInRectangle(ci, cj: Integer; var
  startingPoint: Vector2)
  : Boolean;
var
  position: array[1..9] of Vector2;
  hereStarts: Vector2;
  i, j, k, theBest: Integer;
  f0, badness, leastBadness: double;
begin
  result := FALSE;
  leastBadness := 1.0E100;
  theBest := 1;
  k := 0;
  for i := 1 to 3 do
    for j := 1 to 3 do begin
      k := k + 1;
      position[k, 1] := coordVert[1, ci] + step[1] * coordRatios[i];
      position[k, 2] := coordVert[2, cj] + step[2] * coordRatios[j];
      theF(f0, position[k]);
      badness := abs(f0);
      if (badness < leastBadness) then begin
        leastBadness := badness;
        theBest := k;
      end;
    end; {of the j loop}
  hereStarts := position[thebest];
  if getNearestZero(position[thebest], startingPoint, True, badness,
    hereStarts) then begin
    result := TRUE;
    startingPoint := hereStarts;
  end;
end; {InitialPointInRectangle}



procedure TImplicitCurve.ComputeImplicitCurve(theF0: R2ToR1Procedure;
  InRange0: CheckRange2DFunction;
  RecordAndDraw0: RecordAndDrawProcedure2D; verySmallF0: Double;
  minCorner0, maxCorner0: Vector2; ppLine0, flagsPLine0: Integer;
  limitSegments0: Boolean);
var
  i, j, u1, u2: Integer;
  initialPointFound, inDomain: Boolean;
  pInitial: Vector2;
begin
  SetEnvironment(theF0, InRange0, RecordAndDraw0, verySmallF0, minCorner0,
    maxCorner0, ppLine0, flagsPLine0, limitSegments0);
  FillChar(alreadyUsed, SizeOf(alreadyUsed), #$0);
  pInitial[1] := 0;
  pInitial[2] := 0;
  for i := 0 to ppLine - 1 do
    for j := 0 to ppLine - 1 do begin
      initialPointFound := InitialPointInRectangle(i, j, pInitial);
      if initialPointFound then begin
        u1 := Floor((pInitial[1] - minCorner[1]) / flagStep[1]);
        u2 := Floor((pInitial[2] - minCorner[2]) / flagStep[2]);
        inDomain := (u1 >= 0) and (u1 < flagsPLine) and (u2 >= 0) and
          (u2 < flagsPLine);
        if inDomain and (not alreadyUsed[u1, u2]) then
          FollowLine(pInitial, u1, u2);
      end;
    end;
end;


initialization
  InitializeTImplicitCurve;

end.
