

{-----------------------------------------------------------------------------
 Unit Name: FRunKut
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/01/02 
 Purpose: class TRunKut implements a Runge-Kutta algorithm for ordinary
    differential equations in dimensions two and three and in orders one
    and two.
-----------------------------------------------------------------------------}
{*-
 * Copyright (c) 2002 Angel Montesinos Amilibia
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *        Version 4.0.0.0
 *}



unit FRunKut;

interface

uses
  SysUtils, Classes
{$IFDEF MSWINDOWS}
  , Windows, Forms, Graphics,
{$ENDIF}
{$IFDEF LINUX}
  , QGraphics, QForms, QTypes,
{$ENDIF}
  Globals;
const
  kNothingSpecial = 0;
  kDoubled = 1;
  kTopNotReached = 2;
  kJumpedOdd = 4;
  kMinimumReached = 8;

type
  DerivativeFunction = function(var deriv: Vector3; prevDerivOrVel, here:
    Vector3;
    axis: Integer): Boolean;
  CheckRangeFunction = function(lastX, nextX: Vector3; findProlong: Boolean;
    var newX, newVel: Vector3): Integer;
    {If proX is in parameter range and clip range, result = 1;
    else if it can be consistently jumped to a new starting pos, result = 2
    and newX, newVel are changed to the new values; else result = 0 }
  CheckCameBackFunction = function(firstX, prevX, curX: Vector3; counter:
    Integer): Boolean;
  RecordAndDrawProcedure = procedure(x1, x2: Vector3; ToSave: Boolean);
  AdjustPointProcedure = procedure(var x, vel: Vector3);

  TRunKut = class(TObject)
  private
    Derivative: DerivativeFunction;
    InRange: CheckRangeFunction;
    RecordAndDraw: RecordAndDrawProcedure;
    CheckCameBack: CheckCameBackFunction;
    AdjustPoint: AdjustPointProcedure;
    dim, order: Integer;
    axis: Integer;
    x, target, vel, rk1, rk2, rk3, newX, tempX, tempVel, newVel: Vector3;
    tol, error, milliTol, tStep, tGuessStep, shift, minShift: Double;
    function Initialize(x1, x2, vel0: Vector3; axis0: Integer): Boolean;
    function ComputeCoeffients1: Boolean;
    function ComputeCoeffients2: Boolean;
    function RungeKutta: Integer;
    {This function uses the current values of  't', 'tGuessStep', 'x', and
     'vel' to give the values of 'rk1, rk2, rk3' and 'tStep', so that its
     caller can compute, by interpolation, any value of 'x(t + theta*tStep)',
     where 0<= theta <= 1, with a relative precision less than 'tol', and in
     this case the function result is 1. if at a distance 'tGuessStep' less
     than gRadius2D or gRadius3D 'x(t+tGuessStep)' would be out of the parameter
     domain and the curve is prolongueable, it will give 2 as result; else if it
     finds an obstacle in the required range at a distance less than 'minShift',
     it gives 0 as result, so that the caller must interrupt the computation.}
    function Interpolate1(theta: double): Vector3;
     {Uses 'x', 'tStep' and 'rk1, rk2, rk3' to get an approximation to
    ' x(t+ theta * tStep)'. }
    function Interpolate2(theta: double; var veloc: vector3): vector3;
  public
    t, topT, distToLeft, radius: Double;
    procedure SetEnvironment(dim0, order0: Integer; tol0: Double;
      derivFunction0: DerivativeFunction;
      CheckRange0: CheckRangeFunction; RecordAndDraw0: RecordAndDrawProcedure;
      CheckCameBack0: CheckCameBackFunction; AdjustPoint0:
      AdjustPointProcedure);
    function ComputeIntegralCurve(var x0: Vector3; vel0: Vector3; axis0:
      Integer): Integer;
    function GeodTowardsPoint(x0, x1, vel0: Vector3): Integer;
  end;

var
  theRunKut: TRunKut;

implementation

uses
  Math, Messages,
  FSurface, Utilities, ImplicitUtils;



procedure TRunKut.SetEnvironment(dim0, order0: Integer; tol0: Double;
  derivFunction0: DerivativeFunction;
  CheckRange0: CheckRangeFunction; RecordAndDraw0: RecordAndDrawProcedure;
  CheckCameBack0: CheckCameBackFunction; AdjustPoint0: AdjustPointProcedure);
begin
  Derivative := derivFunction0;
  InRange := CheckRange0;
  RecordAndDraw := RecordAndDraw0;
  CheckCameBack := CheckCameBack0;
  AdjustPoint := AdjustPoint0;
  dim := dim0;
  order := order0;
  tol := tol0;
  milliTol := tol / 1000;
  distToLeft := 1.0E20;
  if dim = 2 then
    radius := gRadius2D
  else
    radius := gRadius3D;
  minShift := radius / 20;
  if (order = 2) and (gGeomOption = ExponentialMap) or (gGeomOption =
    GeodesicBall) then
    topT := 1
  else
    topT := 1.0E20;
end;

function TRunKut.Initialize(x1, x2, vel0: Vector3; axis0: Integer): Boolean;
var
  celerity: Double;
  i: Integer;
begin
  x := x1;
  target := x2;
  vel := vel0;
  if dim = 2 then begin
    x[3] := 0;
    target[3] := 0;
    vel[3] := 0;
  end;
  axis := axis0;
  result := Derivative(rk1, vel, x, axis);
  if result then begin
    celerity := tol;
    if order = 1 then
      for i := 1 to dim do
        celerity := celerity + sqr(rk1[i])
    else
      for i := 1 to dim do
        celerity := celerity + sqr(vel[i]);
    celerity := sqrt(celerity);
    tGuessStep := radius / celerity;
  end;
end;

function TRunKut.ComputeCoeffients1: Boolean;
var
  pos: Vector3;
  i: Integer;
begin
  result := Derivative(rk1, vel, x, axis);
  if result then begin
    for i := 1 to dim do
      pos[i] := x[i] + tGuessStep * rk1[i];
    result := Derivative(rk2, vel, pos, axis);
    if result then begin
      for i := 1 to dim do
        pos[i] := x[i] + 0.25 * tGuessStep * (rk1[i] + rk2[i]);
      result := Derivative(rk3, vel, pos, axis);
      if result then begin
        error := milliTol;
        for i := 1 to dim do
          error := error + abs(tGuessStep * (2 * rk3[i] - rk1[i] - rk2[i]));
        if dim = 2 then
          error := error / gSpan2D
        else
          error := error / gSpan3D;
      end;
    end;
  end;
end;

function TRunKut.ComputeCoeffients2: Boolean;
  {Computes the values of 'rk1, rk2, rk3' from which RunKut2 estimates the}
  {next point of the curve, and also the relative error of the method, 'err'}
var
  provPos, provVel: Vector3;
  i: Integer;
begin
  result := Derivative(rk1, vel, x, axis);
  if result then begin
    for i := 1 to dim do begin
      provPos[i] := x[i] + tGuessStep * vel[i];
      provVel[i] := vel[i] + tGuessStep * rk1[i]
    end;
    result := Derivative(rk2, provVel, provPos, axis);
    if result then begin
      for i := 1 to dim do begin
        provPos[i] := x[i] + 0.25 * tGuessStep * (vel[i] + provVel[i]);
        provVel[i] := vel[i] + 0.25 * tGuessStep * (rk1[i] + rk2[i])
      end;
      result := Derivative(rk3, provVel, provPos, axis);
      if result then begin
        error := milliTol;
        for i := 1 to dim do begin
          error := error + abs(kOneSixth * sqr(tGuessStep) * (rk2[i] -
            rk1[i]));
          error := error + abs(kOneThird * tGuessStep * (2.0 * rk3[i] - rk2[i]
            -
            rk1[i]));
        end;
        if dim = 2 then
          error := error / gSpan2D
        else
          error := error / gSpan3D;
        result := TRUE;
      end;
    end;
  end;
end;

function TRunKut.RungeKutta: Integer;
var
  factor: Double;
  displacement: Vector3;
  i: Integer;
  goodCoeff, findProlong: Boolean;
begin
  result := 0;
  repeat
    if order = 1 then
      goodCoeff := ComputeCoeffients1
    else
      goodCoeff := ComputeCoeffients2;
    if not goodCoeff then
      break;
    shift := 0;
    if order = 1 then begin
      tempVel := vel;
      for i := 1 to dim do begin
        displacement[i] := tGuessStep * kOneSixth * (rk1[i] + rk2[i] + 4.0 *
          rk3[i]);
        tempX[i] := x[i] + displacement[i];
        shift := shift + sqr(displacement[i]);
      end;
    end
    else
      for i := 1 to dim do begin
        displacement[i] := tGuessStep * (vel[i] + tGuessStep * kOneSixth * (2
          * rk1[i] + rk2[i]));
        tempX[i] := x[i] + displacement[i];
        tempVel[i] := vel[i] + tGuessStep * kOneSixth * (rk1[i] + rk2[i] + 4.0
          * rk3[i]);
        shift := shift + sqr(displacement[i]);
      end;

    shift := sqrt(shift);
    if (shift < minShift) then
      break;
    if (error < tol) then begin
      findProlong := (shift < 3 * radius);
      newX := x;
      newVel := tempVel;
      result := InRange(x, tempX, findProlong, newX, newVel);
      if (result = 0) then begin
        if findProlong then
          break
        else
          tGuessStep := tGuessStep / 2;
      end
      else if (t + tGuessStep > topT) then begin
        result := 0;
        tGuessStep := tGuessStep / 2;
      end;
    end
    else begin
      factor := Max(0.5, 0.8 * exp(ln(tol / error) / 3));
      factor := Min(2, factor);
      tGuessStep := factor * tGuessStep;
    end;
  until (result > 0);
  if result > 0 then begin
    tStep := tGuessStep;
    factor := Max(0.3, 0.8 * exp(ln(tol / error) / 3));
    factor := Min(3, factor);
    tGuessStep := factor * tGuessStep;
    if order = 1 then
      vel := rk3;
  end;
end;

function TRunKut.Interpolate1(theta: double): Vector3;
var
  theta2, b1, b2, b3: Double;
  i: Integer;
begin
  theta2 := sqr(theta);
  b1 := theta + (-kThreeOver2 + kTwoOver3 * theta) * theta2;
  b2 := (-0.5 + kTwoOver3 * theta) * theta2;
  b3 := (2 - kFourOver3 * theta) * theta2;
  for i := 1 to dim do
    result[i] := x[i] + tStep * (b1 * rk1[i] + b2 * rk2[i] + b3 * rk3[i]);
end;

function TRunKut.Interpolate2(theta: double; var
  veloc: vector3): vector3;
var
  theta2, b1, b2, b3: double;
  i: integer;
begin
  theta2 := sqr(theta);
  b1 := theta + (-kThreeOver2 + kTwoOver3 * theta) * theta2;
  b2 := (-0.5 + kTwoOver3 * theta) * theta2;
  b3 := (2 - kFourOver3 * theta) * theta2;
  for i := 1 to dim do begin
    result[i] := x[i] + tStep * (b1 * vel[i] + b2 * (vel[i] + tStep * rk1[i])
      +
      b3 * (vel[i] + 0.25 * tStep * (rk1[i] + rk2[i])));
    veloc[i] := vel[i] + tStep * (b1 * rk1[i] + b2 * rk2[i] + b3 * rk3[i]);
  end;
end;

function TRunKut.ComputeIntegralCurve(var x0: Vector3; vel0: Vector3;
  axis0: Integer): Integer;
var
  signal, pieces, i, counter: Integer;
  theta: Double;
  pos1, pos2, vel2: Vector3;
begin
  result := kNothingSpecial;
  if not Initialize(x0, pos1, vel0, axis0) then
    Exit;
  if Assigned(RecordAndDraw) then RecordAndDraw(x, x, True);
  counter := 0;
  pos2 := x;
  t := 0;
  repeat
    if ((counter mod 25) = 0) then
      Application.ProcessMessages;
    if gStop then
      Break;
    inc(counter);
    if Assigned(AdjustPoint) then
      AdjustPoint(x, vel);
    signal := RungeKutta;
    if signal > 0 then begin
      if dim = 2 then
        pieces := round(1 + shift / gMaxSegmentLength2D)
      else
        pieces := round(1 + shift / gMaxSegmentLength3D);
      theta := 1 / pieces;
      if pieces > 1 then begin
        for i := 1 to pieces - 1 do begin
          pos1 := pos2;
          t := t + theta * tStep;
          if order = 1 then
            pos2 := Interpolate1(i * theta)
          else
            pos2 := Interpolate2(i * theta, vel2);
          if Assigned(RecordAndDraw) then RecordAndDraw(pos1, pos2, True);
          if Assigned(CheckCameBack) and CheckCameBack(x0, pos1, pos2, counter)
            then begin
            result := result or kDoubled;
            Exit;
          end;
        end;
      end;
      pos1 := pos2;
      pos2 := tempX;
      if order = 2 then
        vel := tempVel;
      t := t + theta * tStep;
      if Assigned(RecordAndDraw) then RecordAndDraw(pos1, pos2, True);
      if Assigned(CheckCameBack) and CheckCameBack(x0, pos1, pos2, counter)
        then begin
        Result := result or kDoubled;
        Exit;
      end;
      if signal = 1 then
        x := pos2
      else begin
        x := newX;
        pos2 := x;
        vel := newVel;
        result := result xor kJumpedOdd;
      end;
    end;
  until signal = 0;
  x0 := pos2;
  if (order = 2) then begin
    if dim = 2 then
      theta := hypot(vel[1], vel[2])
    else
      theta := EuclNorm3(vel);
    theta := (topT - t) * theta;
    if (theta > radius) then
      result := result or kTopNotReached
  end;

end;

function TRunKut.GeodTowardsPoint(x0, x1, vel0: Vector3): Integer;
var
  signal, pieces, i, counter, numMinima: Integer;
  theta, dist0, dist1, dist2, t1, tMin: Double;
  pos1, pos2, vel1, vel2: Vector3;
begin
  result := kNothingSpecial;
  if not Initialize(x0, x1, vel0, 0) then
    Exit;
  counter := 0;
  pos2 := x;
  t := 0;
  t1 := 0;
  vel1 := vel;
  dist2 := EuclDistance3(x, target);
  dist1 := dist2;
  distToLeft := dist2;
  numMinima := 0;
  repeat
    if ((counter mod 25) = 0) then
      Application.ProcessMessages;
    if gStop then
      Break;
    inc(counter);
    if Assigned(AdjustPoint) then
      AdjustPoint(x, vel);
    signal := RungeKutta;
    if signal > 0 then begin
      if dim = 2 then
        pieces := round(1 + shift / gMaxSegmentLength2D)
      else
        pieces := round(1 + shift / gMaxSegmentLength3D);
      theta := 1 / pieces;
      for i := 1 to pieces do begin
        pos1 := pos2;
        dist0 := dist1;
        dist1 := dist2;
        t := t + theta * tStep;
        pos2 := Interpolate2(i * theta, vel2);
        dist2 := EuclDistance3(target, pos2);
        if dist1 < abs(distToLeft) then begin
          distToLeft := dist1;
          if ((dim = 3) and (CheckHandImplicit(pos1, target, vel1) < 0)) or
            ((dim = 2) and (vel1[1] * (target[2] - pos1[2]) - vel1[2] *
            (target[1] - pos1[1]) < 0)) then
            distToLeft := -distToLeft;
          tMin := t1;
        end;
        if (dist0 > dist1) and (dist1 < dist2) then begin
          inc(numMinima);
          if (numMinima > 1) or (dist1 < radius) then begin
            signal := 0;
            break;
          end;
        end;
        t1 := t;
        vel1 := vel2;
      end;
      vel := tempVel;
      if signal = 1 then
        x := tempX
      else if (signal = 2) then begin
        x := newX;
        pos1 := x;
        vel := newVel;
        result := result xor kJumpedOdd;
      end;
    end;
  until signal = 0;
  if (numMinima > 0) or (abs(distToLeft) < radius) then begin
    t := tMin;
    result := result or kMinimumReached;
  end;
end;


procedure InitRunKut;
begin
  theRunKut := TRunKut.Create;
end;

initialization
  initRunKut;

finalization
  ;
end.
