

{-----------------------------------------------------------------------------
 Unit Name: OCFunct
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/01/02 
 Purpose: Class TOpCodeFunction. Compiles, in direct X86 opcodes for evaluation,
    a function given as a string. It is at least as fast as the same string
    compiled by Delphi at compile time. WARNING: some functions are extended by
    continuity for avoiding errors. For instance, the square root of a
    negative number gives 0; division by a dangerously small number or by
    zero gives an enormous number but not a NAN or an error.
      Some functions that are used along the project are here declared, for
    instance the map of a parametric surface is TheParamMap.
 History: I am indebted to Dmitry Kaschenko and Vladimir Safin for the general
    idea of translating an evaluation tree to direct X86 opcode that I learned
    by studying their component "FormulaCompiler".
-----------------------------------------------------------------------------}
{*-
 * Copyright (c) 2002 Angel Montesinos Amilibia
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *        Version 4.0.0.0
 *}



unit OCFunct;

interface
uses Globals;

type

  TOpCodeFunction = class(TObject)
  private
    rawText: AnsiString; {the text of the function}
    data: array of double; {stores the constants that appear in the formula}
    functionCode: AnsiString; {the function opCode}
    procedure AddIntToCode(n: Integer);
   // PROCEDURE PushToAppStack;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer);
    procedure PopFromAppStack;
    procedure CodifyFunction;
  public
    V: function(const X: array of Double): Double;
    constructor Create;
    destructor Destroy; override;
    function ParseNewText(t, vars: AnsiString): AnsiString;
  end;

function ReadConstantField(theStr: AnsiString; var val: double):
  AnsiString;


procedure TheParamMap(var result: Vector3; u, v: double);
procedure TheParamMap2(var result: Vector3; here: vector2);
procedure TheCurve(var result: Vector3; t: double);

function EuclideanGradOfF(position: Vector3): Vector3;
function MinkowskianGradOfF(position: Vector3): Vector3;

var
  gTheF1, gTheF2, gTheF3, gTheF4: TOpCodeFunction;

implementation
uses
  SysUtils, Math;

const
  pi: double = 3.1415926535897932384;
  halfPi: double = 1.5707963267948966192313;
  twoPi: double = 6.283185307179586476925;
  recipTwoPi: double = 0.1591549430918953357689;
  recipPi: double = 0.3183098861837906715378;
  kMaxNumOfNames = 42; {The number of strings of length less}
                             {than 8 that the parser recognizes as}
                             {function names. For instance, asin  }
                             {and arcsin are both recognized.     }
  upCaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  variables = 'PQRSTUVWXYZ';
  digits = '0123456789';

  eThereIsNoError = 0;
  eBadNumber = 1; {A number is miswritten.}
  eBadCharacters = 2; {You have used an illegal character.}
  eBadFunctionOrVariables = 3; {You mispelled a function or variable}
  eBadParentheses = 4; {Check the parentheses, please.}
  eTooMuchConstants = 5; {I cannot digest that much constants.}
  eSintaxError = 6; {There is a sintax fault in the formula.}
  eErrorInParsing = 7; {I did fail in analysing the function.}
  eExpressionTooLarge = 8; {The expression is too long}
  eNothingWritten = 9; {There was nothing to work on!}
  eBadVar = 10; {The allowed variables in this formula are:}
  eBadVarSet = 11; {The allowed variables must be in the range 'p' to 'z'}

  eMaxNum = 1.0E30;
  eMinNum = 1.0E-30;
  eMaxExp = 100;


  {char constants}
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';
  kLastChar = #$FF;

  {Code constants}

  c_zero_comp = {#$D9#$EE#$DF#$F1} #$50#$D9#$EE#$D8#$D9#$DF#$E0#$9B#$9E#$58;
    {ST(7) must be empty. Compares 0 with x=ST(0). Then, JA jumps
     if 0>x; JNA jumps if 0 <= x; JB, if 0 < x; JNB, if 0 >= x; JNE if 0<>x}
  c_1_comp = {#$D9#$E8#$DF#$F1} #$50#$D9#$E8#$D8#$D9#$DF#$E0#$9B#$9E#$58;
    {ST(7) must be empty. Compares 1 with x=ST(0). Then, JA jumps
     if 1>x; JNA jumps if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}

  c_mod_pi = #$D9#$EB#$D9#$C9#$D9#$F8#$DD#$D9;
    {8 bytes. ST(7) must be empty}

  c_ln = #$D9#$E8#$D9#$C9#$D9#$F1#$D9#$EA#$DE#$F9;
    {10 bytes. Requires ST(0) > 0 and ST(7) empty}
  c_2toX = #$D9#$C0#$D9#$FC#$D9#$C9#$D8#$E1#$D9#$F0#$D9#$E8#$DE#$C1#$D9#$C9#$D9
    +
    #$E8#$D9#$FD#$DD#$D9#$DE#$C9;
    {24 bytes. Requires ST6) and ST(7) empty}

type

  pieceClass = (openParent, closeParent, operationSymb, {}
    negation, constantSymb, varSymb, functionSymb, badThing);

  WrittenFunction = record
    theText: AnsiString;
    equivChar: char;
  end;



var
  uKnownFunc: array[1..kMaxNumOfNames] of WrittenFunction;
  uFormula: AnsiString;
  uAttribution: array['#'..#$9F] of pieceClass;
  uReduced: AnsiString;
  uSpecialValues: array[0..7] of double;
  dataLast: byte;
  uStdMessages: array[1..11] of AnsiString;
  allowedVars: AnsiString; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]}
  argNum: Integer; {Length[allowedVars]}
  actorList: AnsiString; {the RPN string that represents the function}
  lengthActorList: Integer;

  emptyRegisters, waitingInStack: Byte;
  c_mod_2Pi, c_div_S, c_ln_S, c_2toX_S: AnsiString; {code vars}

constructor TOpCodeFunction.Create;
begin
  inherited Create;
  rawText := '';
  functionCode := '';
  data := nil;
end;

destructor TOpCodeFunction.Destroy;
begin
  if self <> nil then begin
    rawText := '';
    functionCode := '';
    data := nil;
  end;
  inherited Destroy;
end;

function inNumber(targetStr: AnsiString;
  var startChar: char;
  var startPos: Integer): double;
var
  valid, alreadyBlank, alreadyDot, foundOther: boolean;
  charPos, stringLength: Integer;
  theChar, oldDecimalSeparator: char;
  numStr: AnsiString;
  theNumber: double;
begin
  oldDecimalSeparator := DecimalSeparator;
  DecimalSeparator := '.';
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  alreadyBlank := false;
  foundOther := False;
  numStr := Concat('', startChar);
  charPos := startPos;
  theChar := startChar;
  theNumber := 1.0;
  while (charPos < stringLength) and valid and (not foundOther) do begin
    inc(charPos);
    theChar := targetStr[charPos];
    case theChar of
      '0', '1', '2', '3', '4', '5', '6', '7', '8', '9':
        if not alreadyBlank then
          numStr := numStr + theChar
        else begin
          valid := true;
          theNumber := theNumber * StrToFloat(numStr);
          numStr := theChar;
          alreadyDot := false;
          alreadyBlank := false;
        end;
      '.':
        if alreadyDot or alreadyBlank then
          valid := False
        else begin
          alreadyDot := True;
          numStr := numStr + theChar;
        end;
      ' ':
        if numStr[length(numStr)] = '.' then
          valid := False
        else
          alreadyBlank := true;
    else begin
        foundOther := true;
        charPos := charPos - 1;
        theChar := targetStr[charPos];
      end;
    end; {de CASE}
  end;
  if valid then begin
    inNumber := theNumber * StrToFloat(numStr);
    startPos := charPos;
    startChar := theChar;
  end
  else
    startChar := '?';
  DecimalSeparator := oldDecimalSeparator;
end; {de inNumber}

function UnifyChars: Integer;
var
  place, depth, anInteger, theLength: Integer;
  currentChar, precChar, lastAdded, toAdd: Char;
begin {UnifyChars}
  UnifyChars := eThereIsNoError;
  theLength := Length(uFormula);
  uFormula := UpperCase(uFormula);
  uReduced := '';
  if theLength = 0 then begin
    UnifyChars := eNothingWritten;
    Exit;
  end;
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    place := place + 1;
    currentChar := uFormula[place];
    case currentChar of
      'A'..'Z': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        end;
      '0'..'9': begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            uReduced := Concat(uReduced, '*');
          toAdd := CurrentChar;
        end;
      '(', '[', '{': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            uReduced := Concat(uReduced, '*');
          toAdd := '(';
          depth := depth + 1;
        end;
      '*':
        if (lastAdded = '*') then begin
          anInteger := Length(uReduced);
          delete(uReduced, anInteger, 1);
          uReduced := concat(uReduced, '^');
          lastAdded := '^';
          toAdd := ' ';
        end
        else
          toAdd := '*';
      ')', ']', '}': begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.', '''':
        toAdd := '.';
      '-':
        if (lastAdded = '(') then
          toAdd := '#'
        else
          toAdd := '-';
      '/', '\':
        toAdd := '/';
      ' ', '^':
        toAdd := currentChar;
      '+':
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
      kReturn, kTab, kLineFeed:
        toAdd := ' ';
    else begin
        UnifyChars := eBadCharacters;
        Exit;
      end;
    end; {of CASE}

    if toAdd <> ' ' then begin
      uReduced := concat(uReduced, toAdd);
      lastAdded := toAdd;
    end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    UnifyChars := eBadParentheses
  else if (place < theLength) then
    UnifyChars := eExpressionTooLarge;
end; {of UnifyChars}

function IdentifyFunction(theFunctionText: AnsiString): char;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                        }
var
  i: Integer;
  foundFunction: Boolean;
begin
  i := 0;
  repeat
    i := i + 1;
    foundFunction := (theFunctionText = uKnownFunc[i].theText);
  until (i = kMaxNumOfNames) or foundFunction;
  if foundFunction then
    IdentifyFunction := uKnownFunc[i].equivChar
  else
    IdentifyFunction := '?';
end;

function FactorizeText(targetString: AnsiString; {}
  var mayEndInFunction: Boolean): AnsiString;
var
  numberOfChars, hotSpot, count, position: Integer;
  currentChar: char;
  presumedText: AnsiString;
  resultingString: AnsiString;
begin {FactorizeText}
  FactorizeText := '';
  resultingString := '';
  numberOfChars := Length(targetString);
  hotSpot := numberOfChars;
  if mayEndInFunction and (numberOfChars > 1) then begin
    count := numberOfChars;
    hotSpot := 1;
    repeat
      presumedText := Copy(targetString, hotSpot, count);
      currentChar := IdentifyFunction(presumedText);
      hotSpot := hotSpot + 1;
      count := count - 1;
    until (currentChar <> '?') or (count = 0);
    if (currentChar <> '?') then begin
      hotSpot := hotSpot - 2;
      mayEndInFunction := TRUE;
      if (hotSpot > 0) then
        resultingString := Concat('*', currentChar)
      else begin
        resultingString := currentChar;
        FactorizeText := resultingString;
        Exit;
      end;
    end
    else begin
      hotSpot := numberOfChars;
      mayEndInFunction := FALSE;
    end;
  end {of IF mayEndInFunction ...}
  else
    mayEndInFunction := FALSE;
  if hotSpot > 0 then begin
    currentChar := targetString[hotSpot];
    position := pos(currentChar, allowedVars);
    if (position <> 0) then
      resultingString := Concat(kLastChar, chr(position), resultingString)
    else begin {check if maybe it is a 'pi'}
      if (currentChar = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
        = 'P') then begin
        hotSpot := hotSpot - 1;
        resultingString := 'p' + resultingString;
      end
      else {fatal error}
        Exit;
    end;
    hotSpot := hotSpot - 1;
  end;
  while hotSpot > 0 do begin
    currentChar := targetString[hotSpot];
    position := pos(currentChar, allowedVars);
    if (position <> 0) then
      resultingString := Concat(kLastChar, chr(position), '*',
        resultingString)
    else begin {check if maybe it is a 'pi'}
      if (currentChar = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
        = 'P') then begin
        hotSpot := hotSpot - 1;
        resultingString := 'p*' + resultingString;
      end
      else {fatal error}
        Exit;
    end;
    hotSpot := hotSpot - 1;
  end; {of WHILE...}
  FactorizeText := resultingString;
end; {FactorizeText}


function ReduceSymbols(aF: TOpCodeFunction): Integer;
var
  numberOfChars, hotSpot, reducedLength: Integer;
  currentChar: char;
  letterStr: AnsiString;
  convertedStr, resultingString: AnsiString;
  numberBegins, inLetter, mayHaveFunction, foundEqualNumber: Boolean;
  realValue: double;
  aByte: Byte;
begin {ReduceSymbols}
  reduceSymbols := eThereIsNoError;
  resultingString := '';
  reducedLength := Length(uReduced);
  numberOfChars := reducedLength;
  hotSpot := 0;
  repeat
    hotSpot := hotSpot + 1;
    currentChar := uReduced[hotSpot];
    inLetter := (pos(currentChar, upCaseLetters) <> 0);
    numberBegins := (pos(currentChar, digits) <> 0) or (currentChar = '.');
    if (not inLetter) and (not numberBegins) then
      ResultingString := Concat(resultingString, currentChar);
    if numberBegins then begin
      realValue := 0;
      if currentChar = '.' then
        currentChar := '?'
      else
        realValue := inNumber(uReduced, currentChar, hotSpot);
      if (currentChar <> '?') and (currentChar <> '.') then begin
        if realValue = 1 then
          resultingString := resultingString + '1'
        else begin
          foundEqualNumber := False;
          if dataLast > 0 then
            for aByte := 1 to dataLast do
              if aF.data[aByte] = realValue then begin
                foundEqualNumber := True;
                break;
              end;
          if foundEqualNumber then
            resultingString := Concat(resultingString, kLastChar, chr(aByte
              +
              argNum))
          else begin
            inc(dataLast);
            if (dataLast + argNum = $FF - 1) then begin
              ReduceSymbols := eTooMuchConstants;
              Exit;
            end;
            if Length(aF.data) <= dataLast then
              SetLength(aF.data, dataLast + 10);
            aF.data[dataLast] := realValue;
            resultingString := Concat(resultingString, kLastChar,
              chr(dataLast + argNum))
          end;
        end
      end
      else begin
        ReduceSymbols := eBadNumber;
        Exit;
      end;
    end; {of numberBegins}
    if inLetter then begin
      letterStr := currentChar;
      while (hotSpot < numberOfChars) and inLetter do begin
        hotSpot := hotSpot + 1;
        currentChar := uReduced[hotSpot];
        inLetter := (pos(currentChar, upCaseLetters) <> 0);
        if inLetter then
          letterStr := Concat(letterStr, currentChar);
      end;
      mayHaveFunction := (currentChar = '(');
      if not inLetter then
        hotSpot := hotSpot - 1;
      convertedStr := FactorizeText(letterStr, mayHaveFunction);
      if convertedStr = '' then begin
        ReduceSymbols := eBadFunctionOrVariables;
        Exit;
      end;
      resultingString := Concat(resultingString, convertedStr);
      if (currentChar = '(') and (not mayHaveFunction) then
        ResultingString := Concat(resultingString, '*');
    end; {of inLetter}

  until (hotSpot >= numberOfChars);
  uReduced := resultingString;
end; {ReduceSymbols}

function CheckFunction(allowedIndepVars: AnsiString): Integer;
{Checks the sintaxis of the function}
var
  lastRead, reducedLength: integer;
  species, speciesAnt: pieceClass;
  charac: char;
  stopHere: boolean;

  function hasError: boolean;
  var
    foundError, usual: boolean;
  begin
    foundError := false;
    if species = badThing then
      foundError := true
    else begin
      usual := (species = operationSymb) or (species = closeParent);
      case speciesAnt of
        openParent:
          foundError := usual;
        closeParent, constantSymb, varSymb:
          foundError := not usual;
        negation, operationSymb:
          foundError := usual or (species = negation);
        functionSymb:
          foundError := (species <> openParent);
      end; {of case}
    end; {of else}
    hasError := foundError;
  end; {of hasError}

begin {CheckFunction}
  CheckFunction := eThereIsNoError;
  lastRead := 0;
  reducedLength := Length(uReduced);
  speciesAnt := openParent;
  repeat
    lastRead := lastRead + 1;
    charac := uReduced[lastRead];
    if charac = kLastChar then begin
      inc(lastRead);
      species := constantSymb;
    end
    else
      species := uAttribution[charac];
    stopHere := hasError;
    speciesAnt := species;
  until (lastRead = reducedLength) or stopHere;
  if (not stopHere) then begin
    species := closeParent;
    stopHere := hasError;
  end;
  if stopHere then
    CheckFunction := eSintaxError;
end; {CheckFunction}

function ConvertPowerSymbol(x: AnsiString; n: Integer): AnsiString;
var
  z: AnsiString;

  function ConvertPowerSymbolProv(xx: AnsiString; nn: Integer): AnsiString;
  var
    zz: AnsiString;
    m, np: Integer;
  begin
    zz := xx;
    np := abs(nn);
    m := np div 2;
    if np = 0 then
      Result := '1'
    else if nn = 1 then
      Result := xx
    else if m > 0 then begin
      result := ConvertPowerSymbolProv(zz, m);
      result := 'D(' + result + ')';
      if odd(np) then
        result := result + '*' + zz;
    end
    else
      Result := zz;
  end;

begin
  z := ConvertPowerSymbolProv(x, n);
  Result := '(' + z + ')';
  if n < 0 then
    Result := '(' + '1/' + Result + ')';
end;

function PieceEnd(const first: Integer; const theText, soughtChars:
  AnsiString; var symbol: char): Integer;
{This function acts upon 'theText'. It is assumed that 'theText[first]'
is not in the set [+,-,*,/,^] (but possibly is '#'). Then, it scans forward
'theText' until in reaches one of the symbols in 'endChars' while being at
>=zero parentheses depth. Result is length[theText] if not found or the index
previous to that symbol.}
var
  i, len, depth: Integer;
  ch: char;
  found: Boolean;
begin
  symbol := '?';
  len := Length(theText);
  i := first - 1;
  depth := 0;
  found := False;
  while (i < len) and not found do begin
    inc(i);
    ch := theText[i];
    if (ch = '(') then
      inc(depth)
    else if (ch = ')') then
      dec(depth);
    if ((pos(ch, soughtChars) <> 0) and (depth = 0)) or (depth < 0) then
      found := true;
  end;
  if found then begin
    PieceEnd := i - 1;
    symbol := ch;
  end
  else
    PieceEnd := len;
end;

function PieceBeginning(const last: Integer; soughtChars: AnsiString):
  Integer;
{This function acts upon 'uReduced'. It is assumed that 'uReduced[last]'
is not in the set [+,-,*,/,^]. Then, it scans backwards
'uReduced' until in reaches one of the symbols in 'endChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
var
  i, depth: Integer;
  ch: char;
  found: Boolean;
begin
  i := last + 1;
  depth := 0;
  found := False;
  while (i > 1) and not found do begin
    dec(i);
    ch := uReduced[i];
    if (ch = ')') then
      inc(depth)
    else if (ch = '(') then
      dec(depth);
    if ((pos(ch, soughtChars) <> 0) and (depth = 0)) or (depth < 0) then
      found := true;
  end;
  if found then
    PieceBeginning := i + 1
  else
    PieceBeginning := 1;
end;

function computeExponent(var n: Integer; aF: TOpCodeFunction; ini, fini:
  Integer): Boolean;
var
  c1, c2: char;
  negate: boolean;
  firstExp, lastExp, shift: Integer;
  exponent: double;
begin
  firstExp := ini;
  lastExp := fini;
  c1 := uReduced[firstExp];
  c2 := uReduced[lastExp];
  if (c1 = '(') then begin
    result := (c2 = ')');
    if result then begin
      inc(firstExp);
      dec(lastExp);
      c1 := uReduced[firstExp];
      c2 := uReduced[lastExp];
    end;
  end;
  negate := (c1 = '#');
  if negate then begin
    inc(firstExp);
    c1 := uReduced[firstExp];
  end;

  if (lastExp = firstExp) and (c1 = '1') then begin
    n := 1;
    Result := True;
  end
  else if ((lastExp - firstExp) = 1) and (c1 = kLastChar) and (ord(c2) >
    argNum)
    then begin
    shift := ord(c2) - argNum;
    exponent := aF.data[shift];
    n := round(exponent);
    Result := (n = exponent);
  end
  else
    result := False;
  if result and negate then n := -n;
end;


procedure convertPowers(aF: TOpCodeFunction);
var
  newString: AnsiString;
  reducedLength, firstBase, lastBase, firstExp, lastExp, i, n: Integer;
  c: char;
  good: Boolean;
begin
  reducedLength := Length(uReduced);
  if reducedLength < 3 then
    Exit;
  i := 1;
  repeat
    inc(i);
    c := uReduced[i];
    if (c = '^') then begin
      firstExp := i + 1;
      lastExp := PieceEnd(firstExp, uReduced, '+-*/^', c);
      good := computeExponent(n, af, firstExp, lastExp);
      if good then begin
        lastBase := i - 1;
        firstBase := PieceBeginning(lastBase, '+-*/^#');
        newString := Copy(uReduced, firstBase, lastBase - firstBase + 1);
        newString := ConvertPowerSymbol(newString, n);
        Insert(newString, uReduced, lastExp + 1);
        Delete(uReduced, firstBase, lastExp - firstBase + 1);
        reducedLength := Length(uReduced);
        i := firstBase + Length(newString) - 1;
      end
    end;
  until (i > reducedLength - 1);
end;



function ExpressionString(theText: AnsiString): AnsiString;
  forward;

function AtomString(theText: AnsiString): AnsiString;
var
  base: AnsiString;
  len: Integer;
  ch: char;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else begin
    ch := theText[1];
    if ch = '(' then begin
      base := Copy(theText, 2, len - 2);
      result := ExpressionString(base);
    end
    else begin
      Assert(not (pos(ch, 'qrstuvwxyz') <> 0));
      case ch of
        '1', 'p':
          Result := ch;
        kLastChar:
          Result := theText;
        'A'..'Z': begin
            base := Copy(theText, 3, len - 3);
            result := ExpressionString(base) + ch;
          end;
      else
        ;
      end;
    end;
  end;
end; {AtomString}

function FactorString(theText: AnsiString): AnsiString;
var
  neg: Boolean;
  redStr, anAtom: AnsiString;
  len, first, last: Integer;
  symbol, prevSymbol: char;
begin
  if (theText = '') then
    Result := ''
  else begin
    neg := (theText[1] = '#');
    if neg then begin
      len := Length(theText) - 1;
      redStr := Copy(theText, 2, len);
    end
    else begin
      len := Length(theText);
      redStr := theText;
    end;
    if len = 0 then
      Result := ''
    else begin
      last := PieceEnd(1, redStr, '^', symbol);
      if last = len then
        Result := AtomString(redStr)
      else begin
        anAtom := Copy(redStr, 1, last);
        prevSymbol := symbol;
        result := AtomString(anAtom);
        repeat
          first := last + 2;
          last := PieceEnd(first, redStr, '^', symbol);
          anAtom := Copy(redStr, first, last - first + 1);
          Result := AtomString(anAtom) + Result + prevSymbol;
          prevSymbol := symbol;
        until last = len;
      end;
      if neg then Result := Result + '#';
    end;
  end;
end; {of FactorString}

function TermString(theText: AnsiString): AnsiString;
var
  aFactor: AnsiString;
  len, first, last: Integer;
  symbol, prevSymbol: char;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else begin
    last := PieceEnd(1, theText, '*/', symbol);
    if last = len then
      Result := FactorString(theText)
    else begin
      aFactor := Copy(theText, 1, last);
      prevSymbol := symbol;
      result := FactorString(aFactor);
      repeat
        first := last + 2;
        last := PieceEnd(first, theText, '*/', symbol);
        aFactor := Copy(theText, first, last - first + 1);
        Result := Result + FactorString(aFactor) + prevSymbol;
        prevSymbol := symbol;
      until last = len;
    end;
  end;
end; {of TermString}


function ExpressionString(theText: AnsiString): AnsiString;
var
  aTerm: AnsiString;
  len, first, last: Integer;
  symbol, prevSymbol: char;
begin
  len := Length(theText);
  if len = 0 then
    Result := ''
  else begin
    last := PieceEnd(1, theText, '+-', symbol);
    if last = len then
      Result := TermString(theText)
    else begin
      aTerm := Copy(theText, 1, last);
      prevSymbol := symbol;
      result := TermString(aTerm);
      repeat
        first := last + 2;
        last := PieceEnd(first, theText, '+-', symbol);
        aTerm := Copy(theText, first, last - first + 1);
        Result := Result + TermString(aTerm) + prevSymbol;
        prevSymbol := symbol;
      until last = len;
    end;
  end;
end; {of ExpressionString}




function TOpCodeFunction.ParseNewText(t, vars: AnsiString): AnsiString;
var
  theErrorNumber, i: Integer;
  ch: char;
begin
  Result := '';
  rawText := t;
  uFormula := t;
  if vars = '' then begin
    allowedVars := variables;
    argNum := Length(variables);
  end
  else begin
    allowedVars := vars;
    StrUpper(PChar(allowedVars));
    argNum := Length(vars);
    for i := 1 to argNum do begin
      ch := allowedVars[i];
      if not (pos(ch, variables) <> 0) then begin
        Result := uStdMessages[eBadVarSet];
        Exit;
      end;
    end;
  end;
  data := nil;
  SetLength(data, 8);
  {data[0] is not used}
  data[0] := 0;
  dataLast := 0;
  theErrorNumber := UnifyChars;
  if theErrorNumber = eThereIsNoError then begin
    theErrorNumber := ReduceSymbols(self);
    if theErrorNumber = eThereIsNoError then begin
      theErrorNumber := CheckFunction(allowedVars);
      if theErrorNumber = eThereIsNoError then begin
        ConvertPowers(self);
        actorList := expressionString(uReduced);
        lengthActorList := Length(actorList);
        CodifyFunction;
        @V := PChar(functionCode);
      end;
    end;
  end;
  if theErrorNumber <> eThereIsNoError then
    Result := uStdMessages[theErrorNumber];
end;

procedure TOpCodeFunction.AddIntToCode(n: Integer);
var
  c: PChar;
  s: AnsiString;
  i: Integer;
begin
  c := @n;
  s := '';
  for i := 0 to SizeOf(Integer) - 1 do
    s := s + c[i];
  functionCode := functionCode + s;
end;

function sn(n: Pointer): AnsiString;
var
  c: PChar;
  pn, i: Integer;
begin
  pn := Integer(n);
  c := @pn;
  Result := '';
  for i := 0 to SizeOf(Integer) - 1 do
    Result := Result + c[i];
end;


procedure TOpCodeFunction.LeaveRoomInFPUStack(neededPlaces: Integer);
var
  i, m: Integer;
begin
  if emptyRegisters >= neededPlaces then
    Exit;
  m := neededPlaces - emptyRegisters;
  if emptyRegisters > 0 then
    for i := 1 to emptyRegisters do
      functionCode := functionCode + #$D9#$F6;
  for i := 1 to m do begin
    inc(waitingInStack);
    functionCode := functionCode + #$D9#$F6#$DD#$9C#$24;
    AddIntToCode(-8 - SizeOf(double) * waitingInStack);
    functionCode := functionCode + #$D9#$F6;
  end;
  for i := 1 to neededPlaces do
    functionCode := functionCode + #$D9#$F7;
  emptyRegisters := neededPlaces;
end;

procedure TOpCodeFunction.PopFromAppStack;
{Simplified, because it is only used when emptyRegisters = 7}
begin
  functionCode := functionCode + #$DD#$84#$24;
  AddIntToCode(-8 - SizeOf(double) * waitingInStack);
  functionCode := functionCode + #$D9#$C9;
  dec(waitingInStack);
  dec(emptyRegisters);
end;

procedure TOpCodeFunction.CodifyFunction;
var
  i, n: Integer;
  ch: char;
  escape: Boolean;
begin
  functionCode := '';
  emptyRegisters := 8;
  waitingInStack := 0;
  i := 0;
  while i < lengthActorList do begin
    inc(i);
    ch := actorList[i];
    escape := (ch = kLastChar);
    if escape or (ch = 'p') or (ch = '1') then begin
      if escape then inc(i);
      ch := actorList[i];
      LeaveRoomInFPUStack(1);
      if escape then begin
        if (ord(ch) > argNum) then begin
          n := Integer(@data[ord(ch) - argNum]);
          functionCode := functionCode + #$DD#$05;
          AddIntToCode(n);
        end
        else begin
          n := ord(ch);
          if n = 1 then
            functionCode := functionCode + #$DD#$00
          else
            functionCode := functionCode + #$DD#$40 + chr(8 * (n - 1));
        end;
      end
      else if (ch = 'p') then
        functionCode := functionCode + #$D9#$EB
      else
        functionCode := functionCode + #$D9#$E8;
      dec(emptyRegisters);
    end
    else begin
      if (ch in ['+', '-', '*', '/', '^']) then begin
        if (emptyRegisters = 7) then
          PopFromAppStack;
        case ch of
          '+':
            functionCode := functionCode + #$DE#$C1;
          '-':
            functionCode := functionCode + #$DE#$E9;
          '*':
            functionCode := functionCode + #$DE#$C9;
          '/': begin
              LeaveRoomInFPUStack(1);
              functionCode := functionCode + c_div_S;
            end;
          '^': begin
              LeaveRoomInFPUStack(1);
              functionCode := functionCode + c_zero_comp +
                #$72#$08#$DD#$D8#$DD#$D8#$D9#$EE#$EB#$52#$D9#$F1 +
                c_2toX_S;
            end;
        end;
        inc(emptyRegisters);
      end
      else {opCode of a function} begin
        if (ch in ['A', 'B', 'C', 'E', 'F', 'G', 'M']) then
          LeaveRoomInFPUStack(1)
        else if (ch in ['H', 'I', 'J', 'K', 'L', 'O', 'P', 'Q', 'R']) then
          LeaveRoomInFPUStack(2);
        case ch of
          '#':
            functionCode := functionCode + #$D9#$E0;
          'A':
            functionCode := functionCode + c_mod_2pi + #$D9#$FF;
          'B':
            functionCode := functionCode + { c_mod_2pi +} #$D9#$FE;
          'C':
            functionCode := functionCode + #$D9#$EA#$DE#$C9 +
              c_2ToX_S;
          'D':
            functionCode := functionCode + #$D8#$C8;
          'E':
            functionCode := functionCode + c_zero_comp +
              #$72#$06 +
              #$DD#$D8#$D9#$EE#$EB#$02#$D9#$FA;
          'F':
            functionCode := functionCode + c_ln_S;
          'G':
            functionCode := functionCode + c_mod_pi + #$D9#$F2#$DD#$D8;

          'H':
            functionCode := functionCode + #$D9#$EA#$DE#$C9 +
              c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
              #$F1#$DE#$E9#$DD#$05 + sn(@uSpecialValues[7]) + #$DE#$C9;
          'I':
            functionCode := functionCode + #$D9#$EA#$DE#$C9 +
              c_2ToX_S + #$D9#$C0#$D9#$E8#$DE +
              #$F1#$DE#$C1#$DD#$05 + sn(@uSpecialValues[7]) + #$DE#$C9;
          'J':
            functionCode := functionCode + #$D9#$EA#$DE#$C9 +
              c_2ToX_S + #$D8#$C8#$D9#$E8#$D8#$E9#$D9#$C9 +
              #$D9#$E8#$DE#$C1#$DE#$F9;
          'K':
            functionCode := functionCode + #$D9#$C0#$D9#$E1 +
              c_1_comp +
              #$DD#$D8#$72#$0E#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$F3 +
              #$EB#$16 + c_zero_comp + #$DD#$D8#$DD#$05 +
              sn(@uSpecialValues[0]) +
              #$72#$02#$D9#$E0;
          'L':
            functionCode := functionCode + #$D9#$C0#$D9#$E1 +
              c_1_comp +
              #$DD#$D8#$72#$10#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E1#$D9#$FA#$D9#$C9 +
              #$D9#$F3#$EB#$14 + c_zero_comp +
              #$DD#$D8#$72#$04#$D9#$EB#$EB#$02 +
              #$D9#$EE;
          'M': functionCode := functionCode + #$D9#$E8#$D9#$F3;
          'N': functionCode := functionCode + #$D9#$E1;
          'O':
            functionCode := functionCode + c_mod_pi +
              #$D9#$FB#$D9#$C9 + c_div_S;
          'P':
            functionCode := functionCode + c_1_comp +
              #$77#$18#$D9#$C0#$D8#$C8#$D9#$E8#$DE#$E9#$D9#$FA#$DE#$C1 +
              c_ln + #$EB#$04#$DD#$D8#$D9#$EE;
          'Q':
            functionCode := functionCode +
              #$D9#$C0#$D9#$E1 + c_1_comp + #$DD#$D8#$76#$22
              +
              #$D9#$C0#$D9#$E8#$DE#$C1#$D9#$C9#$D9#$E8#$DE#$E1#$DE#$F9 +
              c_ln + #$DD#$05 + sn(@uSpecialValues[7]) + #$DE#$C9#$EB#$1C
              + c_zero_comp +
              #$DD#$D8#$77#$08#$DD#$05 + sn(@uSpecialValues[2]) +
              #$EB#$06#$DD#$05 + sn(@uSpecialValues[3]);
          'R':
            functionCode := functionCode + #$D9#$C0#$D8#$C8#$D9#$E8#$DE#$C1
              +
              #$D9#$FA#$DE#$C1 + c_ln;
        else
          ;
        end;
      end;
    end;
  end; {del while}
  functionCode := functionCode + #$C3;
end;



procedure InitializeCodeVars;
begin
  c_mod_2Pi := #$DD#$05 + sn(@uSpecialValues[1]) + #$D9#$C9#$D9#$F8#$DD#$D9;
  c_div_S := c_zero_comp + #$75#$0C#$DD#$D8#$DD#$05 + sn(@uSpecialValues[2]) +
    #$DE#$C9#$EB#$02#$DE#$F9; {Safe division. 20 bytes. Requires ST(7) empty}
  c_ln_S := c_zero_comp + #$72#$0C#$DD#$D8#$DD#$05 + sn(@uSpecialValues[6]) +
    #$D9#$E0#$EB#$0A + c_ln;
    {Safe ln(ST(0)). 26 bytes. Requires ST(7) empty}
  c_2toX_S := #$D9#$C0#$D9#$E1#$DD#$05 + sn(@uSpecialValues[6]) + #$DE#$E9 +
    c_zero_comp +
    #$DD#$D8#$72#$1A + c_2toX + #$EB#$1C + c_zero_comp +
      #$DD#$D8#$72#$08#$DD#$05
    +
    sn(@uSpecialValues[4]) + #$EB#$06#$DD#$05 + sn(@uSpecialValues[2]);
     {Safe 2^ST(0). 72 bytes. Requires ST(6) and ST(7) empty.}
end;



procedure InitializeParser;
var
  character: char;
begin
  uKnownFunc[1].theText := 'COS';
  uKnownFunc[1].equivChar := 'A';

  uKnownFunc[2].theText := 'SIN';
  uKnownFunc[2].equivChar := 'B';

  uKnownFunc[3].theText := 'SEN';
  uKnownFunc[3].equivChar := 'B';

  uKnownFunc[4].theText := 'EXP';
  uKnownFunc[4].equivChar := 'C';

  uKnownFunc[5].theText := 'SQR';
  uKnownFunc[5].equivChar := 'D';

  uKnownFunc[6].theText := 'SQRT';
  uKnownFunc[6].equivChar := 'E';

  uKnownFunc[7].theText := 'LN';
  uKnownFunc[7].equivChar := 'F';

  uKnownFunc[8].theText := 'LOG';
  uKnownFunc[8].equivChar := 'F';

  uKnownFunc[9].theText := 'TG';
  uKnownFunc[9].equivChar := 'G';

  uKnownFunc[10].theText := 'TAN';
  uKnownFunc[10].equivChar := 'G';

  uKnownFunc[11].theText := 'SH';
  uKnownFunc[11].equivChar := 'H';

  uKnownFunc[12].theText := 'SINH';
  uKnownFunc[12].equivChar := 'H';

  uKnownFunc[13].theText := 'CH';
  uKnownFunc[13].equivChar := 'I';

  uKnownFunc[14].theText := 'COSH';
  uKnownFunc[14].equivChar := 'I';

  uKnownFunc[15].theText := 'TH';
  uKnownFunc[15].equivChar := 'J';

  uKnownFunc[16].theText := 'TANH';
  uKnownFunc[16].equivChar := 'J';

  uKnownFunc[17].theText := 'ARCSEN';
  uKnownFunc[17].equivChar := 'K';

  uKnownFunc[18].theText := 'ASEN';
  uKnownFunc[18].equivChar := 'K';

  uKnownFunc[19].theText := 'ARCSIN';
  uKnownFunc[19].equivChar := 'K';

  uKnownFunc[20].theText := 'ASIN';
  uKnownFunc[20].equivChar := 'K';

  uKnownFunc[21].theText := 'ARCCOS';
  uKnownFunc[21].equivChar := 'L';

  uKnownFunc[22].theText := 'ACOS';
  uKnownFunc[22].equivChar := 'L';

  uKnownFunc[23].theText := 'ARCOS';
  uKnownFunc[23].equivChar := 'L';

  uKnownFunc[24].theText := 'ARCTG';
  uKnownFunc[24].equivChar := 'M';

  uKnownFunc[25].theText := 'ARCTAN';
  uKnownFunc[25].equivChar := 'M';

  uKnownFunc[26].theText := 'ATAN';
  uKnownFunc[26].equivChar := 'M';

  uKnownFunc[27].theText := 'ABS';
  uKnownFunc[27].equivChar := 'N';

  uKnownFunc[28].theText := 'COTG';
  uKnownFunc[28].equivChar := 'O';

  uKnownFunc[29].theText := 'COTAN';
  uKnownFunc[29].equivChar := 'O';

  uKnownFunc[30].theText := 'COT';
  uKnownFunc[30].equivChar := 'O';

  uKnownFunc[31].theText := 'ARGCH';
  uKnownFunc[31].equivChar := 'P';

  uKnownFunc[32].theText := 'ACOSH';
  uKnownFunc[32].equivChar := 'P';

  uKnownFunc[33].theText := 'ARCCOSH';
  uKnownFunc[33].equivChar := 'P';

  uKnownFunc[34].theText := 'ARCTANH';
  uKnownFunc[34].equivChar := 'Q';

  uKnownFunc[35].theText := 'ARGTANH';
  uKnownFunc[35].equivChar := 'Q';

  uKnownFunc[36].theText := 'ARGTH';
  uKnownFunc[36].equivChar := 'Q';

  uKnownFunc[37].theText := 'ATANH';
  uKnownFunc[37].equivChar := 'Q';

  uKnownFunc[38].theText := 'ARCSINH';
  uKnownFunc[38].equivChar := 'R';

  uKnownFunc[39].theText := 'ASINH';
  uKnownFunc[39].equivChar := 'R';

  uKnownFunc[40].theText := 'ARGSH';
  uKnownFunc[40].equivChar := 'R';

  uKnownFunc[41].theText := 'ARGSINH';
  uKnownFunc[41].equivChar := 'R';

  uKnownFunc[42].theText := 'ARGCOSH';
  uKnownFunc[42].equivChar := 'P';

  for character := '#' to 'z' do
    case character of
      '(':
        uAttribution[character] := openParent;
      '#':
        uAttribution[character] := negation;
      ')':
        uAttribution[character] := closeParent;
      '+', '-', '*', '/', '^':
        uAttribution[character] := operationSymb;
      '1', 'p':
        uAttribution[character] := constantSymb;
      'q'..'z':
        uAttribution[character] := varSymb;
      'A'..'Z':
        uAttribution[character] := functionSymb;
    else
      uAttribution[character] := badThing;
    end; {of case}

  {Translate these messages below to your language. Their meaning}
 {in English is given in the IMPLEMENTATION CONST part.         }

  uStdMessages[1] := 'There is a miswritten number.';
  uStdMessages[2] := 'You did use a character not allowed.';
  uStdMessages[3] := 'Some function or variable was miswritten.';
  uStdMessages[4] := 'Check, please, the parentheses.';
  uStdMessages[5] := 'I cannot cope with so much constants.';
  uStdMessages[6] := 'There is a sintax defect in the formula.';
  uStdMessages[7] := 'Sorry: I cannot understand that formula.';
  uStdMessages[8] := 'This is an expression too long.';
  uStdMessages[9] := 'Perhaps you forgot to write something.';
  uStdMessages[10] := 'The independent variables must be: ';
  uStdMessages[11] :=
    'The allowed variables must be in the range ''p'' to ''z''';

  uSpecialValues[0] := halfPi;
  uSpecialValues[1] := twoPi;
  uSpecialValues[2] := eMaxNum;
  uSpecialValues[3] := -eMaxNum;
  uSpecialValues[4] := eMinNum;
  uSpecialValues[5] := -eMinNum;
  uSpecialValues[6] := eMaxExp;
  uSpecialValues[7] := 0.5;
  InitializeCodeVars;
  gTheF1 := TOpCodeFunction.Create;
  gTheF2 := TOpCodeFunction.Create;
  gTheF3 := TOpCodeFunction.Create;
  gTheF4 := TOpCodeFunction.Create;
end; {InitializeParser}


function ReadConstantField(theStr: AnsiString; var val: double):
  AnsiString;
var
  forConstFunct: TOpCodeFunction;
  warning: AnsiString;
begin
  val := 0;
  forConstFunct := TOpCodeFunction.create;
  warning := forConstFunct.ParseNewText(theStr, '');
  if warning = '' then
    val := forConstFunct.V([]);
  ReadConstantField := warning;
  forConstFunct.Free;
end;

procedure theCurve(var result: Vector3; t: double);
begin
  Result[1] := gTheF1.V([t]);
  Result[2] := gTheF2.V([t]);
  Result[3] := gTheF3.V([t]);
end;

procedure TheParamMap(var result: Vector3; u, v: double);
begin
  Result[1] := gTheF1.V([u, v]);
  Result[2] := gTheF2.V([u, v]);
  Result[3] := gTheF3.V([u, v]);
end;

procedure TheParamMap2(var result: Vector3; here: vector2);
begin
  Result[1] := gTheF1.V(here);
  Result[2] := gTheF2.V(here);
  Result[3] := gTheF3.V(here);
end;


function MinkowskianGradOfF(position: Vector3): Vector3;
var
  place: Vector3;
  fu, fmu, fv, fmv, fw, fmw: double;
begin
  place := position;
  place[1] := position[1] + gIncr3D[1];
  fu := gTheF1.V(place);
  place[1] := position[1] - gIncr3D[1];
  fmu := gTheF1.V(place);
  place[1] := position[1];
  place[2] := position[2] + gIncr3D[2];
  fv := gTheF1.V(place);
  place[2] := position[2] - gIncr3D[2];
  fmv := gTheF1.V(place);
  place[2] := position[2];
  place[3] := position[3] + gIncr3D[3];
  fw := gTheF1.V(place);
  place[3] := position[3] - gIncr3D[3];
  fmw := gTheF1.V(place);
  MinkowskianGradOfF[1] := 0.5 * (fu - fmu) / gIncr3D[1];
  MinkowskianGradOfF[2] := 0.5 * (fv - fmv) / gIncr3D[2];
  MinkowskianGradOfF[3] := 0.5 * (fmw - fw) / gIncr3D[3];
end;

function EuclideanGradOfF(position: Vector3): Vector3;
var
  place: Vector3;
  fu, fmu, fv, fmv, fw, fmw: double;
begin
  place := position;
  place[1] := position[1] + gIncr3D[1];
  fu := gTheF1.V(place);
  place[1] := position[1] - gIncr3D[1];
  fmu := gTheF1.V(place);
  place[1] := position[1];
  place[2] := position[2] + gIncr3D[2];
  fv := gTheF1.V(place);
  place[2] := position[2] - gIncr3D[2];
  fmv := gTheF1.V(place);
  place[2] := position[2];
  place[3] := position[3] + gIncr3D[3];
  fw := gTheF1.V(place);
  place[3] := position[3] - gIncr3D[3];
  fmw := gTheF1.V(place);
  EuclideanGradOfF[1] := 0.5 * (fu - fmu) / gIncr3D[1];
  EuclideanGradOfF[2] := 0.5 * (fv - fmv) / gIncr3D[2];
  EuclideanGradOfF[3] := 0.5 * (fw - fmw) / gIncr3D[3];
end;

procedure FreeFunctions;
begin
  gTheF1.free;
  gTheF2.free;
  gTheF3.free;
  gTheF4.free;
end;

initialization
  InitializeParser;
finalization
  FreeFunctions;

end.


